@@ -1,6 +1,7 @@
 import sys
 import json
 from operator import itemgetter
+import re
 
 import requests
 from http_signature.requests_auth import HTTPSignatureAuth
@@ -31,6 +32,16 @@ DEFAULT_HEADERS = {
 
 DEBUG_CONFIG = {'verbose': sys.stderr}
 
+
+def search_dicts(dicts, predicate, fields):
+    matcher = re.compile(predicate, re.IGNORECASE)
+    for d in dicts:
+        m = [d for f in fields if matcher.search(d.get(f, ''))]
+        if m:
+            yield m[0]
+            continue
+
+
 class DataCenter(object):
     """
     Basic connection object that makes all API requests.
@@ -175,7 +186,7 @@ class DataCenter(object):
             self.auth = HTTPSignatureAuth(key_id=key_id, secret=secret, 
                 allow_agent=allow_agent)
     
-    def request(self, method, path, headers=None, **kwargs):
+    def request(self, method, path, headers=None, data=None, **kwargs):
         """
         (Primarily) internal method for making all requests to the datacenter.
         
@@ -196,9 +207,13 @@ class DataCenter(object):
         request_headers.update(self.default_headers)
         if headers:
             request_headers.update(headers)
+        if data:
+            data = json.dumps(data)
+            kwargs['data'] = data
         resp = requests.request(method, full_path, auth=self.auth, 
             headers=request_headers, config=self.config, **kwargs)
-        if resp.status_code == 401 and self.auth and self.auth.signer._agent_key:
+        if (resp.status_code == 401 and self.auth and 
+                self.auth.signer._agent_key):
             self.auth.signer.swap_keys()
             return self.request(method, path, headers=headers, **kwargs)
         if 400 <= resp.status_code < 499:
@@ -338,18 +353,30 @@ class DataCenter(object):
         dc.auth = self.auth
         return dc
     
-    def datasets(self):
+    def datasets(self, search=None, fields=('description', 'urn')):
         """
         ::
         
             GET /:login/datasets
         
+        :param search: optionally filter (locally) with a regular expression 
+            search on the listed fields
+        :type search: :py:class:`basestring` that compiles as a regular 
+            expression
+        
+        :param fields: filter on the listed fields (defaulting to 
+            ``description`` and ``urn``)
+        :type fields: :py:class:`list` of :py:class:`basestring`\s
+        
         :Returns: datasets (operating system templates) available in this 
             datacenter 
         :rtype: :py:class:`list` of :py:class:`dict`\s
         """
         j, _ = self.request('GET', 'datasets')
-        return j
+        if search:
+            return list(search_dicts(j, search, fields))
+        else:
+            return j
     
     def default_dataset(self):
         """
@@ -358,10 +385,11 @@ class DataCenter(object):
             GET /:login/datasets
         
         :Returns: the default dataset for this datacenter
-        :rtype: :py:class:`dict`
+        :rtype: :py:class:`dict` or ``None``
         
         Requests all the datasets in this datacenter, filters for the default, 
-        and returns the corresponding :py:class:`dict`.
+        and returns the corresponding :py:class:`dict`, if a default has been 
+        defined.
         """
         sets = filter(itemgetter('default'), self.datasets())
         if sets:
@@ -389,18 +417,29 @@ class DataCenter(object):
         j, _ = self.request('GET', 'datasets/' + str(dataset_id))
         return j
     
-    def packages(self):
+    def packages(self, search=None, fields=('name',)):
         """
         ::
         
             GET /:login/packages
         
+        :param search: optionally filter (locally) with a regular expression 
+            search on the listed fields
+        :type search: :py:class:`basestring` that compiles as a regular 
+            expression
+        
+        :param fields: filter on the listed fields (defaulting to ``name``\)
+        :type fields: :py:class:`list` of :py:class:`basestring`\s
+        
         :Returns: packages (machine "sizes", with resource types and values) 
             available in this datacenter.
         :rtype: :py:class:`list` of :py:class:`dict`\s
         """
         j, _ = self.request('GET', 'packages')
-        return j
+        if search:
+            return list(search_dicts(j, search, fields))
+        else:
+            return j
     
     def default_package(self):
         """
@@ -409,10 +448,10 @@ class DataCenter(object):
             GET /:login/packages
         
         :Returns: the default package for this datacenter
-        :rtype: :py:class:`dict`
+        :rtype: :py:class:`dict` or ``None``
         
         Requests all the packages in this datacenter, filters for the default, 
-        and returns the corresponding dict.
+        and returns the corresponding dict, if a default has been defined.
         """
         packages = filter(itemgetter('default'), self.packages())
         if packages:
@@ -440,21 +479,43 @@ class DataCenter(object):
         j, _ = self.request('GET', 'packages/' + str(name))
         return j
     
-    def num_machines(self):
+    def num_machines(self, machine_type=None, dataset=None, state=None, 
+            memory=None, tombstone=None, tags=None):
         """
         ::
         
             HEAD /:login/machines
         
-        :Returns: a count of the number of machines owned by the user at this
-            datacenter
+        Counts the number of machines total, or that match the predicates as
+        with :py:meth:`machines`.
+        
+        :param machine_type: virtualmachine or smartmachine
+        :type machine_type: :py:class:`basestring`
+        
+        :param dataset: unique ID or URN for a dataset
+        :type dataset: :py:class:`basestring` or :py:class:`dict`
+        
+        :param state: current running state
+        :type state: :py:class:`basestring`
+        
+        :param memory: current size of the RAM deployed for the machine (Mb)
+        :type memory: :py:class:`int`
+        
+        :param tombstone: include machines destroyed in the last N minutes
+        :type tombstone: :py:class:`int`
+        
+        :param tags: keys and values to query in the machines' tag space
+        :type tags: :py:class:`dict`
+        
+        :Returns: a count of the number of machines (matching the predicates) 
+            owned by the user at this datacenter
         :rtype: :py:class:`int`
         """
         _, r = self.request('HEAD', 'machines')
         num = r.headers.get('x-resource-count', 0)
         return int(num)
     
-    def raw_machine_data(self, machine_id):
+    def raw_machine_data(self, machine_id, credentials=False):
         """
         ::
         
@@ -462,17 +523,25 @@ class DataCenter(object):
         
         :param machine_id: identifier for the machine instance
         :type machine_id: :py:class:`basestring` or :py:class:`dict`
+        
+        :param credentials: whether the SDC should return machine credentials
+        :type credentials: :py:class:`bool`
+        
         :rtype: :py:class:`dict`
         
         Primarily used internally to get a raw dict for a single machine.
         """
+        params = {}
         if isinstance(machine_id, dict):
             machine_id = machine_id['id']
-        j, _ = self.request('GET', 'machines/' + str(machine_id))
+        if credentials:
+            params['credentials'] = True
+        j, _ = self.request('GET', 'machines/' + str(machine_id), 
+                params=params)
         return j
     
     def machines(self, machine_type=None, name=None, dataset=None, state=None, 
-            memory=None, tombstone=None, tag_dict=None, credentials=False, 
+            memory=None, tombstone=None, tags=None, credentials=False, 
             paged=False, limit=None, offset=None):
         """
         ::
@@ -502,8 +571,8 @@ class DataCenter(object):
         :param tombstone: include machines destroyed in the last N minutes
         :type tombstone: :py:class:`int`
         
-        :param tag_dict: keys and values to query in the machines' tag space
-        :type tag_dict: :py:class:`dict`
+        :param tags: keys and values to query in the machines' tag space
+        :type tags: :py:class:`dict`
         
         :param credentials: whether to include the generated credentials for 
             machines, if present
@@ -539,8 +608,8 @@ class DataCenter(object):
             params['memory'] = memory
         if tombstone:
             params['tombstone'] = tombstone
-        if tag_dict:
-            for k, v in tag_dict.items():
+        if tags:
+            for k, v in tags.items():
                 params['tag.' + str(k)] = v
         if credentials:
             params['credentials'] = True
@@ -569,7 +638,7 @@ class DataCenter(object):
         return [Machine(datacenter=self, data=m) for m in machines]
     
     def create_machine(self, name=None, package=None, dataset=None,
-            metadata_dict=None, tag_dict=None):
+            metadata=None, tags=None, boot_script=None, credentials=False):
         """
         ::
         
@@ -591,13 +660,16 @@ class DataCenter(object):
             unique ID or URN
         :type dataset: :py:class:`basestring` or :py:class:`dict`
         
-        :param metadata_dict: keys & values with arbitrary supplementary 
-            details for the machine
-        :type metadata_dict: :py:class:`dict`
+        :param metadata: keys & values with arbitrary supplementary 
+            details for the machine, accessible from the machine itself
+        :type metadata: :py:class:`dict`
         
-        :param tag_dict: keys & values with arbitrary supplementary 
+        :param tags: keys & values with arbitrary supplementary 
             identifying information for filtering when querying for machines
-        :type tag_dict: :py:class:`dict`
+        :type tags: :py:class:`dict`
+        
+        :param boot_script: path to a file to upload for execution on boot
+        :type boot_script: :py:class:`basestring` as file path
         
         :rtype: :py:class:`smartdc.machine.Machine`
         
@@ -618,16 +690,19 @@ class DataCenter(object):
             if isinstance(dataset, dict):
                 dataset = dataset.get('urn', dataset['id'])
             params['dataset'] = dataset
-        if metadata_dict:
-            for k, v in metadata_dict.items():
+        if metadata:
+            for k, v in metadata.items():
                 params['metadata.' + str(k)] = v
-        if tag_dict:
-            for k, v in tag_dict.items():
+        if tags:
+            for k, v in tags.items():
                 params['tag.' + str(k)] = v
-        j, r = self.request('POST', 'machines', params=params)
+        if boot_script:
+            with open(boot_script) as f:
+                params['metadata.user-script'] = f.read()
+        j, r = self.request('POST', 'machines', data=params)
         return Machine(datacenter=self, data=j)
     
-    def machine(self, machine_id):
+    def machine(self, machine_id, credentials=False):
         """
         ::
         
@@ -643,6 +718,7 @@ class DataCenter(object):
             machine_id = machine_id['id']
         elif isinstance(machine_id, Machine):
             machine_id = machine_id.id
-        return Machine(datacenter=self, machine_id=machine_id)
+        return Machine(datacenter=self, machine_id=machine_id, 
+                credentials=credentials)
     
 
