@@ -43,7 +43,8 @@ class Machine(object):
     cache in most cases, instead requiring the user to explicitly update with 
     a :py:meth:`refresh` call.
     """
-    def __init__(self, datacenter, machine_id=None, data=None):
+    def __init__(self, datacenter, machine_id=None, data=None, 
+            credentials=False):
         """
         :param datacenter: datacenter that contains this machine
         :type datacenter: :py:class:`smartdc.datacenter.DataCenter`
@@ -54,6 +55,9 @@ class Machine(object):
         :param data: raw data for instantiation
         :type data: :py:class:`dict`
         
+        :param credentials: whether credentials should be returned
+        :type credentials: :py:class:`bool`
+        
         Typically, a :py:class:`smartdc.machine.Machine` object is 
         instantiated automatically by a 
         :py:class:`smartdc.datacenter.DataCenter` object, but a user may 
@@ -88,7 +92,8 @@ class Machine(object):
         """the :py:class:`smartdc.datacenter.DataCenter` object that holds 
         this machine"""
         if not data:
-            data = self.datacenter.raw_machine_data(self.id)
+            data = self.datacenter.raw_machine_data(self.id, 
+                    credentials=credentials)
         self._save(data)
     
     def __str__(self):
@@ -132,6 +137,10 @@ class Machine(object):
         self.disk = data.get('disk')
         self.ips = data.get('ips', [])
         self.metadata = data.get('metadata', {})
+        if not hasattr(self, '_credentials'):
+            self._credentials = {}
+        self._credentials.update(self.metadata.pop('credentials', {}))
+        self.boot_script = self.metadata.pop('user-script', None)
         self.created = dt_time(data.get('created'))
         self.updated = dt_time(data.get('updated'))
     
@@ -183,19 +192,39 @@ class Machine(object):
         """
         return datacenter.create_machine(**kwargs)
     
-    def refresh(self):
+    def refresh(self, credentials=False):
         """
         ::
         
             GET /:login/machines/:id
         
+        :param credentials: whether to return machine passwords
+        :type credentials: :py:class:`bool`
+        
         Fetch the existing state and values for the 
         :py:class:`smartdc.machine.Machine` from the datacenter and commit the 
         values locally.
         """
-        data = self.datacenter.raw_machine_data(self.id)
+        data = self.datacenter.raw_machine_data(self.id, 
+                credentials=credentials)
         self._save(data)
     
+    def credentials(self):
+        """
+        ::
+        
+            GET /:login/machines/:id?credentials=True
+        
+        :Returns: known login-password pairs for the machine
+        :rtype: :py:class:`dict`
+        
+        Provisionally re-fetch the machine information with the credentials 
+        flag enabled.
+        """
+        if not self._credentials:
+            self.refresh(credentials=True)
+        return self._credentials
+    
     def status(self):
         """
         ::
@@ -351,7 +380,7 @@ class Machine(object):
         in the :py:attr:`metadata` attribute and returns it.
         """
         j, _ = self.datacenter.request('POST', self.path + '/metadata', 
-                    params=kwargs)
+                    data=kwargs)
         self.metadata = j
         return j
     
@@ -376,6 +405,40 @@ class Machine(object):
         r.raise_for_status()
         return self.get_metadata()
     
+    def set_boot_script(self, filename):
+        """
+        ::
+        
+            POST /:login/machines/:id/metadata
+        
+        :param filename: file path to the script to be uploaded and executed
+            at boot on the machine
+        :type filename: :py:class:`basestring`
+        
+        Replace the existing boot script for the machine with the data in the 
+        named file.
+        """
+        data = {}
+        with open(filename) as f:
+            data['user-script'] = f.read()
+        j, r = self.datacenter.request('POST', self.path + '/metadata', 
+                    data=data)
+        r.raise_for_status()
+        self.boot_script = data['user-script']
+    
+    def delete_boot_script(self):
+        """
+        ::
+        
+            DELETE /:login/machines/:id/metadata/user-script
+        
+        Deletes any existing boot script on the machine.
+        """
+        j, r = self.datacenter.request('DELETE', self.path + 
+                '/metadata/user-script')
+        r.raise_for_status()
+        self.boot_script = None
+    
     def delete_all_metadata(self):
         """
         ::
@@ -417,7 +480,7 @@ class Machine(object):
         already set for the machine.
         """
         j, _ = self.datacenter.request('POST', self.path + '/tags', 
-            params=kwargs)
+            data=kwargs)
         return j
     
     def get_tag(self, tag):
@@ -501,7 +564,7 @@ class Machine(object):
         """
         params = {'name': name}
         j, _ = self.datacenter.request('POST', self.path + '/snapshots', 
-            params=params)
+            data=params)
         return Snapshot(machine=self, data=j, name=name)
     
     def start_from_snapshot(self, name):
